#' Thurstone Case V solution to Paired Comparison Data
#'  with logistic probability
#'
#'  @param f vector of # of times that the left object was preferred.
#'  @param n vector of # of trials per pair  or  scalar .
#'  @param ij matrix indicating the stimulus pair:  i, j.
#'  @param sname a vector consisting of stimulus name \cr
#'  When sname is NULL the elements of ij will be used as the names.
#'  @param minp minimum value of probability when n=1.
#'  @param print = 1 to print result.
#'
#'  @details
#'  f_ij is the # of times that sutimulus i is preferred over stimulus j
#'  out of n_ij comparisons. \cr
#'
#'  f/n will be converted to z by logit function and \cr
#'  z will be converted to the difference of centered scale values x.
#'
#'  @return
#' x The scale value \cr
#' sname Stimulus name \cr
#' n2o Name conversion table \cr
#' llh log likelihood \cr
#' maxag max value of the gradient vector \cr
#' varx The variance of the scale values \cr
#' errorvar The variance of the error term associated with x \cr
#' xs Scale values normalized to have unit variance \cr
#' errorvars The variance of the error term associated with xs \cr
#' xls LS scale values \cr
#' errorvalls The variance of xls \cr
#' iter # of iterations used. \cr
#' eps The convergence criterion \cr
#'
#'
#' @examples
#' set.seed(1701)
#' x <- c(-.5,0,1)
#' ij <- matrix(c(2,1, 3,1, 3,2), 3,2, byrow=1)
#' ijc <- t( apply(4-ij,1, function(x) paste("s",x,sep="")) )
#' z <- c(x[2]-x[1], x[3]-x[1], x[3]-x[2])
#' p <- logistic(z)
#' n <- c(10,10,10)
#' f <- mapply( function(size,prob){ rbinom(1,size,prob) }, n, p )
#' res1 <- c5ls( f, n, ij )
#' res1 <- c5ls( f, n, ijc, sname=c("s33","s22","s11") )
#'
#' @export
#'
c5ls <- function( f, n, ij, sname=NULL, minp=1e-9, print=1 ){
 # Thurstone's Case Five solution to the paired comparison data
 # Shin-ichi Mayekawa
 # 20150418 cot, 20
 # data compression: 20150608
 # error check: 20150608
 # sname: 20160709
 #
 #
 # Args:
 #
 #   (f, n, ij)    triplet of vectors
 #                 f[k] is the # of times that object ij[k,1]
 #                 is prefered to object ij[k,2] out of n[k] comparisons.
 #                 n can be a scalar.
 #
 # Values:
 #
 #    x            vector of centered scale values
 #    varx, errorvar, xs, errorvars
 #    pc, nc, fc, ijc, Gc
 #

 npair=length(f)
 if( length(n) == 1 ) n=rep(n,npair)
 # error check
 if( !is.vector(f) | !is.vector(n) | !is.matrix(ij) ){
  cat("\n\nerror1:(c5ls) f must be a vector, ij must be a matrix. \n\n")
  return()
 }

 # rename ij
 #sname_all=sort(unique(c(ij)))
 #o2n=compress_id( ij )$o2n
 #ij=matrix( o2n[ij],,2 )
 nnn=new_name( ij, sort=2 )
 ij=nnn$new
 n2o=nnn$n2o
 sname_all=n2o$old
 rm(nnn)

 npair=length(f)
 if( length(n) == 1 ) n=rep(n,npair)

 # reflect right and left if left < right.
 loc=which(ij[,1] < ij[,2])
 ij[loc,]=cbind(ij[loc,2],ij[loc,1])
 f[loc]=n[loc]-f[loc]


 # stimulus name
 if( is.null(sname) ){
  # sname=sort(unique(c(ij)))
  sname=sname_all
 }
 else{
  if( length(unique(c(ij))) != length(sname) ){
   cat("\nerror1:(c5ls) The length of sname parameter is too short"
       , " or too long. \n")
  }
 }

 # const
 np=length(sname)
 npairfull=np*(np-1)/2

 if( print > 0 ){
  cat("\n\nThurstone's Case Five Solution\n\n")
  Print( npair, npairfull, np, minp )
 }

 # compress data if possible
 ff=xtabs( f ~ ij[,1]+ij[,2] )
 nn=xtabs( n ~ ij[,1]+ij[,2] )
 rn=as.numeric(rownames(ff)); cn=as.numeric(colnames(ff))
 rn1=as.numeric(rownames(nn)); cn1=as.numeric(colnames(nn))
 # Print(ff,nn,rn,cn)
 if( all(rn == rn1) & all(cn == cn1) ){
  # data can be compressed.
  F=matrix(0,np,np); N=F
  for( i in 1:length(rn) )
   for( j in 1:length(cn) ){
    F[rn[i],cn[j]]=ff[i,j]; N[rn[i],cn[j]]=nn[i,j]
   }
  vF=vech(F,nodiag=1); vN=vech(N,nodiag=1)
  IJ=matrix(0,np*(np-1)/2,2)
  k=0;
  for( i in 2:np ) for( j in 1:(i-1) ){
   k=k+1; IJ[k,1]=i; IJ[k,2]=j
  }
  loc=vN != 0
  vF=vF[loc]; vN=vN[loc]; ij=IJ[loc,]
  p=vF/vN
  npair=length(p)
 }
 else{
  # no compression
  p=f/n
 }


 # prob at each pair
 p[p < minp]=minp
 p[p>1-minp]=1-minp
 z=logit(p)

 # G matrix for  paired comparison
 G=matrix(0,npair,np)
 for( k in 1:npair ){
  G[k,ij[k,1]]=1; G[k,ij[k,2]]=-1
 }
 invGG=matSwp(t(G)%*%G)
 rankG=sum(abs(diag(invGG)) > 1e-7)

 if( print >= 9 ){
  Print( vF,vN,ij,G, fmt=".0")
  Print(p,z)
 }

 # scale value
 x=invGG%*%t(G)%*%z
 x=x-mean(x)
 rownames(x)=sname

 # error variance etc
 varx=var(x)/np*(np-1)
 errorvar=1/2
 errorstd=sqrt(errorvar)
 xs=x/sqrt(c(varx))
 varxs=1
 errorvars=errorvar/varx
 errorstds=sqrt(errorvars)

 if( print > 0 ){
  cat("\n Thurstone's Case Five \n")
  Print( x, xs )
  Print( varx, errorvar, errorstd )
  Print( varxs, errorvars, errorstds )
 }

 res=list( x=x, sname=sname, Gc=G, rankGc=rankG
           , varx=varx, errorvar=errorvar, xs=xs, errorvars=errorvars
           , pc=p, nc=n, fc=f, ijc=ij )
 return( res )

} # end of c5ls



#' ML Solution to Thurstone Case V of Paired Comparison Data
#'  with logistic probability
#'
#'  @param f vector of # of times that the left object was preferred.
#'  @param n vector of # of trials per pair  or  scalar .
#'  @param ij matrix indicating the stimulus pair:  i, j.
#'  When sname is NULL the elements of ij will be used as the names.
#'  @param minp minimum value of probability when n=1.
#'  @param lmax max # of iterations.
#'  @param eps criterion for convergence.
#'  @param print = 1 to print result.
#'  @details
#'  f_ij is the # of times that sutimulus i is preferred over stimulus j
#'  out of n_ij comparisons. \cr
#'  f_ij is assumed to have binomial distribution win n_ij and p_ij \cr
#'  where p_ij = logistic( x_i - x_j ).
#'
#'  @examples
#' set.seed(1701)
#' x <- c(-.5,0,1)
#' ij <- matrix(c(2,1, 3,1, 3,2), 3,2, byrow=1)
#' ijc <- t( apply(4-ij,1, function(x) paste("s",x,sep="")) )
#' z <- c(x[2]-x[1], x[3]-x[1], x[3]-x[2])
#' p <- logistic(z)
#' n <- c(10,10,10)
#' f <- mapply( function(size,prob){ rbinom(1,size,prob) }, n, p )
#' res2 <- c5ml( f, n, ij )
#' res1 <- c5ml( f, n, ijc, sname=c("s33","s22","s11") )
#'
#'  @export
#'

c5ml <- function( f, n, ij, sname=NULL, minp=1e-9, lmax=50, eps=1e-6
                  , print=1 ){
 # ML Solution of Thurstone's Case Five for paired comparison data
 # Shin-ichi Mayekawa
 # 20150418 cot,19,20,23
 # new id when ij does not contain all the pairs: 20150428
 # compress_id -> new_id: 20150608
 # error check: 20150608
 # logistic -> logistic1: 20150629
 # sname: 20160709
 #
 #
 # Args:
 #
 #   (f, n, ij)    triplet of vectors
 #                 f[k] is the # of times that object ij[k,1]
 #                 is prefered to object ij[k,2] out of n[k] comparisons.
 #                 n can be a scalar.
 #
 #                 Elements of ij do not have to be unique or sorted.
 #                 ij can be a subset of all the possible pairs.
 #
 #
 # Values:
 #
 #    x            vector of centered scale values
 #    sname        id of the elements of x according to ij
 #    n2o=n2o
 #    llh, maxag
 #    varx, errorvar
 #    xs, errorvars
 #    xls, errorvarls
 #    iter, eps
 #

 llhc5 <- function( f, n, G, x ){
  # evaluation of log likelihood of c5
  # Shin-ichi Mayekawa
  # 20150419cot
  #
  # Args:
  #
  #  z = G x
  #

  p=logistic1(G%*%x)
  # Print(p, log(p), log(1-p))
  llh=sum(f*log(p)+(n-f)*log(1-p))
  return( llh )

 } # end of llhc5



 npair=length(f)
 if( length(n) == 1 ) n=rep(n,npair)
 # error check
 if( !is.vector(f) | !is.vector(n) | !is.matrix(ij) ){
  cat("\n\nerror1:(c5ml) f must be a vector, ij must be a matrix. \n\n")
  return()
 }

 # rename ij
 #sname_all=sort(unique(c(ij)))
 #o2n=compress_id( ij )$o2n
 #ij=matrix( o2n[ij],,2 )
 nnn=new_name( ij, sort=2 )
 ij=nnn$new
 n2o=nnn$n2o
 sname_all=n2o$old
 nstot=length(unique(c(ij)))
 rm(nnn)


 # reflect right and left if left < right.
 loc=which(ij[,1] < ij[,2])
 ij[loc,]=cbind(ij[loc,2],ij[loc,1])
 f[loc]=n[loc]-f[loc]


 # From here, ij contains numeric ids of nstot stimuli.


 # stimulus name
 if( is.null(sname) ){
  # sname=sort(unique(c(ij)))
  sname=sname_all
 }
 else{
  if( length(unique(c(ij))) != length(sname) ){
   cat("\nerror1:(c5ml) The length of sname parameter is too short"
       , " or too long. \n")
  }
 }

 # const
 np=length(sname)
 npairu=nrow(unique(ij))
 npairfull=np*(np-1)/2
 ntot=sum(n)
 npp=ntot/npairu
 pmiss=1-npairu/npairfull
 # Print(np,npairu,npairfull,ntot,npp)

 if( print > 0 ){
  cat("\n\nML Solution of Thurstone's Case Five\n\n")
  cat(" # of pairs observed =", npair, "\n")
  cat(" # of nuique pairs observed =", npairu, "\n")
  cat(" # of stimuli used=", np, "\n")
  cat(" # of full pairs =", npairfull, ",  proportion missing =", pmiss, "\n")
  cat(" total # of comparisons =", ntot, "\n")
  cat(" average # of trials per pair =", npp, "\n")


  if( print >= 3 ){
   Print( ij,f,n )
  }

  cat("\n max # of iterations =", lmax, ",  eps =", eps, "\n")
  cat(" Minimum value of prob is set equal to", minp, " .\n")
 }


 # case five solution
 resls=c5ls( f, n, ij, sname=sname, minp=minp,  print=0 )
 xls=resls$x; errorvarls=resls$errorvar
 rankG=resls$rankG

 # G matrix for difference:  NOT compressed.
 G=matrix(0,npair,np)
 for( k in 1:npair ){
  G[k,ij[k,1]]=1; G[k,ij[k,2]]=-1
 }

 llh00=llhc5( f, n, G, xls )

 # expand or shrink x0 to find a better initial
 llh0=llh00
 x0=xls
 ddd=1
 for( dd in c(3, 2, 1.5, 0.8, 0.6, 0.4, 0.2, 0.1) ){
  llh=llhc5( f, n, G, dd*x0 )
  if( llh > llh0 ){
   ddd=dd
   llh0=llh
  }
 }
 x0=ddd*x0
 # Print( ddd, llh0, llh00 )

 if( print > 0 ){
  cat("\n Initial scale values \n")
  Print( xls, x0, fmt="9.3" )
  Print( llh00, llh0, ddd, np, rankG )
 }

 # Print( n,f,G, x0)

 llh=llh0
 llhp=llh0
 x=x0

 for( llll in 1:lmax ){

  # first derivative of llh
  z=G%*%x
  p=logistic1(z)
  dllhdx=1.7*t(G)%*%(f-n*p)
  maxag=max(abs(dllhdx))

  # Fisher info mat
  H=1.7^2*t(G)%*%diag(c(n*p*(1-p)))%*%G

  # direction
  d=matSwp(H)%*%dllhdx

  # halving
  step=1
  for( lllll in 1:20 ){
   x1=x+step*d
   llh1=llhc5( f, n, G, x1 )
   #Print(lllll, x1, llh1)
   if( llh1 > llh ) break
   step=step/2
  }
  x=x1
  x=x-mean(x)
  llh=llh1

  llhimp=llh-llhp
  llhimpr=llhimp/abs(llhp)

  if( print >= 2 ){
   Print( llll, lllll, llh, llhimp, maxag )
  }

  if( llhimp <= eps ){
   if( print > 0 ) cat("\n c5ml iteration converged.\n" )
   break
  }

  llhp=llh

 } # end of llll loop


 # error variance etc
 rownames(x)=sname
 varx=var(x)/np*(np-1)
 errorvar=1/2
 errorstd=sqrt(errorvar)
 xs=x/sqrt(c(varx))
 varxs=1
 errorvars=errorvar/varx
 errorstds=sqrt(errorvars)

 names(x)=sname; names(xs)=sname

 if( print > 0 ){
  cat("\nML Solution of Thurstone's Case Five \n")
  Print( llh, maxag )
  Print(  x, xs, fmt="9.3" )
  Print( varx, errorvar, errorstd )
  Print( varxs, errorvars, errorstds )
 }

 res=list( x=x, sname=sname, n2o=n2o, llh=llh, maxag=maxag
           , varx=varx, errorvar=errorvar
           , xs=xs, errorvars=errorvars
           , xls=xls, errorvarls=errorvarls
           , iter=llll, eps=eps )
 return( res )

} # end of c5ml


