#' Maxmul Likelihood Solution to MDPREF
#'
#' @param f vector consisting of the # of times that the left stimuli was
#' chosen out of n trials.
#' @param n vector consisting of the # of trials.
#' @param ij matrix indicating the stimulus pair.
#' @param subj vector indicating the subject.
#' @param ndim # of dimensions
#' @param lmax max # of iterations
#' @param eps criterion for convergence
#' @param minp minimum value of the probability
#' @param lmax2 max # of iterations for the inner iteration for X and Beta
#' @param print = 1 to print the result
#' @param plot = 1 to plot the result
#'
#' @details The quartet
#' ( subj, n, f, ij ) contains the result of the paired comparison data.\cr
#' Thoese four objects has the same length or # of rows.\cr
#' The k-th elements of the quartet indicates that \cr
#' subj[k] preferred stimulus ij[k,1] over ij[k,2] f[k] times when exposed
#' to the pair n[k] times.
#' \cr
#' The paired comparizon for each subject does no have to be complete.\cr
#'
#' @examples
#' ij <- cbind( data_comp$ij.1, data_comp$ij.2)
#' mdprefml( f=data_comp$f, n=data_comp$n, ij=ij, subj=data_comp$subj, plot=1 )
#'
#' ij <- cbind( data_miss$ij.1, data_miss$ij.2)
#' mdprefml( f=data_miss$f, n=data_miss$n, ij=ij, subj=data_miss$subj, plot=1 )
#'
#'
#' @export
#'
mdprefml <- function( f, n, ij, subj, ndim=2, lmax=50, eps=1e-6
                      , minp=1e-7, lmax2=2, print=1, plot=0 ){
 # ML solution to MDPREF
 # Shin-ichi Mayekawa
 # 20150418 cot, 19,20,23,24,25
 # arguments added to estbg and estX: 20150426
 # new id when ij does not contain all the pairs: 20150428,29
 # normalization bug fix: 20150429,30
 # compress_id -> new_name: 20150502
 # group -> subj, np -> nstim: 20150503,04
 # natural sort, subject name: 20150508,09
 # npp corrected: 20150510
 # reflect right and left: 20150523
 # biplot: 20150523
 # error check: 20150608
 # logistic -> logistic1: 20150629
 #
 #
 # Args:
 #
 #   (f, n, ij)    triplet of vectors
 #                 f[k] is the # of times that object ij[k,1]
 #                 is prefered to object ij[k,2] out of n[k] comparisons.
 #                 n can be a scalar.
 #
 #                 Elements of ij do not have to be unique.
 #                 ij can be a subset of all the possible pairs.
 #
 #   subj          vector to indicate subjects membership
 #
 #   ndim          # of dimensions
 #
 #   minp          minimum value of probability
 #   lmax          max # of iterations
 #   eps           eps for relative improvement of log likelihood
 #   lmax2         max # of inner iterations
 #
 #
 #
 # Values:
 #
 #    X            nstot x ndim    matrix of centered scale values
 #    B            nsubj x ndim    matrix of subject vectors
 #    Gi           nstot x nsubj   Gi = X B'  (scale value for each subject)
 #
 #    sname_all    list of sorted unique elements of ij.
 #                 rowname(X)=sname_all
 #
 #         If Xall contains the scale values of the larger set of stimuli
 #         Xall[sname_all,]=X will put the result in it.
 #
 # Needs:
 #
 #  logistic, logit, matSwp, princ, new_name
 #
 #  library(naturalsort)
 #

 llhmdpref <- function( f, n, ij, locsubj, nstot, X, B ){
  # evaluation of log likelihood of mdpref
  # Shin-ichi Mayekawa
  # 20150423
  # subj -> locsubj: 20150509
  # logistic -> logistic1: 20150629
  #
  # Args:
  #

  nsubj=nrow(B)
  ndim=ncol(X)

  llh=0
  for( s in 1:nsubj ){
   # subsetting
   locs=locsubj[[s]]
   ijs=ij[locs,]; fs=f[locs]; ns=n[locs]; npairs=nrow(ijs)
   G=matrix(0,npairs,nstot)
   for( k in 1:npairs ){ G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1 }
   ps=logistic1(G%*%X%*%t(B[s,,drop=0]))
   llh=llh+sum(fs*log(ps)+(ns-fs)*log(1-ps))
  }
  return( llh )

 } # end of llhmdpref


 llhc5 <- function( f, n, G, x ){
  # evaluation of log likelihood of c5
  # Shin-ichi Mayekawa
  # 20150419cot
  # logistic -> logistic1: 20150629
  #
  # Args:
  #
  #  z = G x  and  p = logistic( z )
  #

  p=logistic1(G%*%x)
  # Print(p, log(p), log(1-p))
  llh=sum(f*log(p)+(n-f)*log(1-p))
  return( llh )

 } # end of llhc5


 estbs <- function( fs, ns, GX, bs ){
  # estimate bs given X
  # Shin-ichi Mayelawa
  # 20150426
  # logistic -> logistic1: 20150629
  #

  llhbsp=llhc5( fs, ns, GX, bs )
  llhbs=llhbsp

  # first derivative of llh
  zs=GX%*%bs
  ps=logistic1(zs)
  dllhdb=1.7*t(GX)%*%(fs-ns*ps)
  maxagbs=max(abs(dllhdb))

  # Fisher info mat
  H=1.7^2*t(GX)%*%diag(c(ns*ps*(1-ps)))%*%GX

  # direction
  d=matSwp(H)%*%dllhdb

  # halving
  step=1
  for( lll in 1:20 ){
   bs1=bs+step*d
   llhbs1=llhc5( fs, ns, GX, bs1 )
   #Print(lllll, bs1, llh1)
   if( llhbs1 > llhbs ) break
   step=step/2
  }

  # update bs assuming llhbs1 > llhbs
  bs=bs1
  llhbs=llhbs1

  llhbsimp=llhbs-llhbsp
  llhbsimpr=llhbsimp/abs(llhbsp)

  if( print >= 3 ){
   Print( ll, lll, llhbs, llhbsimp, maxagbs )
  }

  return( list(bs=bs, maxagbs=maxagbs, llhbs=llhbs) )

 } # end of estbs


 estX <- function( f, n, ij, locsubj, X, B ){
  # estimate X given B
  # Shin-ichi Mayelawa
  # 20150426,0506
  # subj -> locsubj: 20150509
  #

  notpt=nrow(X)
  nsubj=nrow(B)
  llhx=llhmdpref( f, n, ij, locsubj, nstot, X, B )
  llhxp=llhx

  x=matrix(X,,1)

  # first derivative and Fisher info mat
  dllhdx=0
  H=0
  for( s in 1:nsubj ){
   # subsetting
   locs=locsubj[[s]]
   ijs=ij[locs,]; fs=f[locs]; ns=n[locs]; npairs=nrow(ijs)
   G=matrix(0,npairs,nstot);
   for( k in 1:npairs ){ G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1 }
   bs=t(B[s,,drop=0]) # this is nr x 1
   ps=logistic1(G%*%X%*%bs)
   ttbsG=t(t(bs)%x%G)
   dllhdx=dllhdx+ttbsG%*%(fs-ns*ps)
   H=H+ttbsG%*%(diag(c(ns*ps*(1-ps)))%*%t(ttbsG))
  } # end of subj
  dllhdx=1.7*dllhdx
  H=(1.7^2)*H
  maxagX=max(abs(dllhdx))

  # direction
  d=matSwp(H)%*%dllhdx

  # halving
  step=1
  for( lll in 1:10 ){
   x1=x+step*d
   X1=matrix(x1,nstot)
   # center X1 before evaluating llh
   X1=scale(X1, center=TRUE, scale=FALSE)
   llhx1=llhmdpref( f, n, ij, locsubj, nstot, X1, B )
   # Print(ll, lll, llhx1)
   if( llhx1 > llhx ) break
   step=step/2
  }

  # update X assuming llhx1 > llhx
  X=X1
  llhx=llhx1

  llhximp=llhx-llhxp
  llhximpr=llhximp/abs(llhxp)

  if( print >= 3 ){
   Print( ll, lll, llhx, llhximp, maxagX )
  }

  return( list(X=X, maxagX=maxagX, llhx=llhx) )

 } # end of estX


 if( length(n) == 1 ) n=rep(n,length(f))

 # error check
 if( !is.vector(f) | !is.vector(n) | !is.matrix(ij) ){
  cat("\n\nerror1:(mdprefml) f must be a vector, ij must be a matrix. \n\n")
  return()
 }


 # rename ij
 #sname_all=sort(unique(c(ij)))
 #o2n=compress_id( ij )$o2n
 #ij=matrix( o2n[ij],,2 )
 nnn=new_name( ij, sort=2 )
 ij=nnn$new

 # reflect right and left if left < right.
 loc=which(ij[,1] < ij[,2])
 ij[loc,]=cbind(ij[loc,2],ij[loc,1])
 f[loc]=n[loc]-f[loc]

 n2o=nnn$n2o
 sname_all=n2o$old
 nstot=length(unique(c(ij)))
 rm(nnn)

 # From here, ij contains numeric ids of nstot stimuli.


 # const
 subjname=unique(subj)     # not sorted, order of appearance
 nsubj=length(subjname)

 npair=numeric(nsubj)
 npairfull=numeric(nsubj)
 nstim=numeric(nsubj)
 npairu=numeric(nsubj)
 ntot=numeric(nsubj)
 npp=numeric(nsubj)
 pmiss=numeric(nsubj)
 sname=NULL
 locsubj=NULL
 locmiss=vector("list",nsubj)

 for( s in 1:nsubj ){
  loc=which(subj == subjname[s])
  locsubj[[s]]=loc
  fs=f[loc]; ns=n[loc]; ijs=ij[loc,]
  npair[s]=length(fs)
  sname[[s]]=sort(unique(c(ijs)))
  if( length(setdiff(1:nstot,sname[[s]])) > 0 ){
   locmiss[[s]]=setdiff(1:nstot,sname[[s]])
  }
  nstim[s]=c(length(sname[[s]]))
  npairu[s]=nrow(unique(ijs))
  npairfull[s]=nstim[s]*(nstim[s]-1)/2
  ntot[s]=sum(ns)
  npp[s]=ntot[s]/npairu[s]
  pmiss[s]=1-npairu[s]/npairfull[s]
 }

 nparam=(nstot-1)*ndim + nsubj*ndim  -  ndim*(ndim+1)/2 -  ndim*(ndim-1)/2

 dimname=paste("d",1:ndim,sep="")

 if( print > 0 ){
  cat("\n\nML Solution of MDPREF\n\n")
  # Print(nsubj, npair, npairu, ns, npairfull, pmiss, ntot, npp, locsubj)
  # Print(sname)
  cat(" total # of stimuli used=",nstot,"\n")
  cat(" # of subjects =", nsubj,"\n")
  cat(" # of dimensions =", ndim,"\n\n")
  cat(" # of pairs observed =", npair, "\n")
  cat(" # of unique pairs observed =", npairu, "\n")
  cat(" # of stimuli =", nstim, "\n")
  cat(" # of full pairs =", npairfull,"\n")
  cat(" proportion missing =", pmiss, "\n")
  cat(" total # of comparisons =", ntot, "\n")
  cat(" # of trials per pair =", npp, "\n")
  cat("\n # of parameters to be estimated =", nparam,"\n")
  cat(" max # of iterations =", lmax, ",  eps =", eps, "\n")
  cat(" max # of inner iterations =", lmax2, ",  eps =", eps, "\n")
  cat(" Minimum value of prob is set equal to", minp, " .\n")
 }

 if( print >= 3 ){
  Print( subj, ij,f,n )
 }


 # initial by case five solution
 Gi=matrix(0,nstot,nsubj)
 for( s in 1:nsubj ){
  G=matrix(0,npair[s],nstot)
  ijs=ij[locsubj[[s]],,drop=0]
  for( k in 1:npair[s] ){
   G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1
  }
  ps=f[locsubj[[s]]]/n[locsubj[[s]]]
  zs=logit1(ps, 0.1)
  # Print(s,ijs, ps, zs)
  x=matSwp(t(G)%*%G)%*%t(G)%*%zs
  x[locmiss[[s]]]=NA
  Gi[,s]=x
 }
 Gi=scale(Gi, center=TRUE, scale=FALSE)

 # expand or shrink x0 to find a better initial
 for( s in 1:nsubj ){
  # subsetting
  locs=locsubj[[s]]
  ijs=ij[locs,,drop=0]; fs=f[locs]; ns=n[locs]; npairs=nrow(ijs)
  G=matrix(0,npairs,nstot);
  for( k in 1:npairs ){ G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1 }

  x0=Gi[,s]
  x0[locmiss[[s]]]=0
  llh00=llhc5( fs, ns, G, x0 )
  llh0=llh00
  ddd=1
  for( dd in c(3, 2, 1.5, 0.8, 0.6, 0.4, 0.2, 0.1, 0.05) ){
   llh=llhc5( fs, ns, G, dd*x0 )
   #  Print(dd,llh0,llh)
   if( llh > llh0 ){
    ddd=dd
    llh0=llh
   }
  }
  Gi[,s]=ddd*x0
  Gi[locmiss[[s]],s]=NA
  if( print >= 2 ) Print( s, ddd, llh0, llh00 )
 }
 # Print(Gi)

 # initial X and B by pca of Gi=X B'
 res=princ( Gi, ndim, estmu=1, maxiter=1000, eps=1e-4, init=2, print=0 )
 X=res$F; B=res$A
 rownames(X)=sname_all
 rownames(B)=subjname

 llh0=llhmdpref( f, n, ij, locsubj, nstot, X, B )
 if( print >= 3 ){
  Print(X,B)
  Print(llh0)
 }


 # main iteration starts here

 llh=llh0
 llhp=llh0

 for( llll in 1:lmax ){

  # update B given X

  maxagB=0
  for( s in 1:nsubj ){
   # subsetting
   locs=locsubj[[s]]
   ijs=ij[locs,]; fs=f[locs]; ns=n[locs]; npairs=nrow(ijs)
   G=matrix(0,npairs,nstot);
   for( k in 1:npairs ){ G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1 }
   bs=B[s,]
   for( ll in 1:lmax2 ){
    res=estbs( fs, ns, G%*%X, bs )
    bs=res$bs
    # Print("B ", ll, bs)
    maxagB=max(maxagB,res$maxagbs)
   }
   B[s,]=bs
  } # end of subj

  llh=llhmdpref( f, n, ij, locsubj, nstot, X, B )
  # Print("after B  ", llh)


  # update X given B

  for( ll in 1:lmax2 ){
   res=estX( f, n, ij, locsubj, X, B )
   X=res$X
   rownames(X)=sname_all
   # Print("X ", ll, X)
  }
  maxagX=res$maxagX


  if( llll%%10 == 0 ){
   # normalize X and B
   # make X orthonormal
   ev=eigen( t(X)%*%X/nstot )
   evamhalf=diag(sqrt(1/ev$values)); eve=ev$vectors
   X=X%*%eve%*%evamhalf
   B=B%*%eve%*%diag(sqrt(ev$values))
   # make B orthogonal
   eve=eigen( t(B)%*%B )$vector
   B=B%*%eve
   X=X%*%eve
   # Print("normalized")
  }


  # check convergence
  llh=llhmdpref( f, n, ij, locsubj, nstot, X, B )
  # Print("after X  ", llh)

  llhimp=llh-llhp
  llhimpr=llhimp/abs(llhp)

  if( print >= 2 ){
   Print( llll, llh, llhimpr, maxagB, maxagX )
  }

  if( llhimpr <= eps ){
   if( print > 0 ) cat("\n mdprefml iteration converged.\n" )
   break
  }

  llhp=llh

 } # end of llll loop


 # normalize X and B so that X'X = nstot I  and B'B = diag.
 # make X orthonormal
 X=scale(X,center=TRUE,scale=FALSE)
 ev=eigen( t(X)%*%X/nstot )
 evamhalf=diag(sqrt(1/ev$values)); eve=ev$vectors
 X=X%*%eve%*%evamhalf
 B=B%*%eve%*%diag(sqrt(ev$values))
 # make B orthogonal
 eve=eigen( t(B)%*%B )$vector
 B=B%*%eve
 X=X%*%eve
 # make B positive
 bs=colSums(B>0) < nsubj/2
 B[,bs]=-B[,bs]
 X[,bs]=-X[,bs]

 rownames(X)=sname_all
 colnames(X)=dimname
 rownames(B)=subjname
 colnames(B)=dimname

 # error variance etc
 Gi=X%*%t(B)
 md=mandd(Gi)
 DispGi=md$cov
 stdGi=md$std
 CorGi=diag(1/stdGi)%*%DispGi%*%diag(1/stdGi)
 dimnames(CorGi)=dimnames(DispGi)
 BB=t(B)%*%B
 rcont=diag(BB)/sum(diag(BB))
 names(rcont)=dimname

 aic=-2*(llh-nparam)
 bic=-2*llh+nparam*log(sum(ntot))

 if( print > 0 ){
  cat("\nML Solution of MDPREF \n")
  Print( llh, maxagX, maxagB )
  Print( nparam, aic, bic )
  cat("\n")
  Print( llh, maxagX, maxagB )
  Print( X, B, rcont, fmt="6.3" )
  Print(Gi, fmt="6.3")
  Print(DispGi, BB, fmt="6.3")
  Print(CorGi, fmt="6.3")
 }

 if( plot >=  1 ){
  title=paste("Biplot:  # of params = ", nparam
              ,",   llh = " ,formatC(llh,8,format="fg")
              ,",   aic = ",formatC(aic,8,format="fg")
              ,",   bic = ", formatC(bic,8,format="fg"), sep="")
  biplot( X,B, main=title)
 }

 res=list( X=X, B=B, sname_all=sname_all, sname=sname, n2o=n2o
           , npair=npair, npairu=npairu
           , llh=llh, maxagX=maxagX, maxagB=maxagB
           , Gi=Gi, DispGi=DispGi, iter=llll, eps=eps  )
 return( res )

} # end of mdprefml



#' Finite Mixture Model for MDPREF
#'
#' @param f vector consisting of the # of times that the left stimuli was
#' chosen out of n trials.
#' @param n vector consisting of the # of trials.
#' @param ij matrix indicating the stimulus pair.
#' @param subj vector indicating the subject.
#' @param ndim # of dimensions
#' @param nclass # of laten classes
#' @param alpha hyper-parameter vector of size nclass for
#' Dirichlet prior of rho
#' @param init method for initial configuration\cr
#' = 0 to use random number \cr
#' = 1 to use Thurstome Case V, PCA and hclust
#' = 2 to use mdprefml, PCA and hclust
#' @param X initial configuration, if any
#' @param B initial configuration, if any
#' @param rho initial configuration, if any
#' @param estX
#' = 0 to skip the estimation of X and fix it to the value given in X.
#' @param estB
#' = 0 to skip the estimation of B and fix it to the value given in B.
#' @param estrho
#'  = 0 to skip the estimation of rho and fix it to the value given in rho.
#' @param lmax max # of iterations
#' @param eps criterion for convergence
#' @param minp minimum value of the probability
#' @param lmax2 max # of iterations for the inner iteration for X and Beta
#' @param print = 1 to print the result
#' @param plot = 1 to plot the result
#'
#' @details The quartet
#' ( subj, n, f, ij ) contains the result of the paired comparison data.\cr
#' Thoese four objects has the same length or # of rows.\cr
#' The k-th elements of the quartet indicates that \cr
#' subj[k] preferred stimulus ij[k,1] over ij[k,2] f[k] times when exposed
#' to the pair n[k] times.
#' \cr
#' The paired comparizon for each subject does no have to be complete.\cr
#'
#' @examples
#' ij <- cbind( data_comp$ij.1, data_comp$ij.2)
#' mdprefmx( f=data_comp$f, n=data_comp$n, ij=ij, subj=data_comp$subj, plot=1 )
#'
#' ij <- cbind( data_miss$ij.1, data_miss$ij.2)
#' mdprefmx( f=data_miss$f, n=data_miss$n, ij=ij, subj=data_miss$subj, plot=1 )
#'
#'
#' @export

 mdprefmx <- function( f, n, ij, subj, ndim=2, nclass=2, alpha=1, init=2
                       , X=NULL, B=NULL, rho=NULL, estX=1, estB=1, estrho=1
                       , minp=1e-7, lmax=10, eps=1e-6, lmax2=3, print=1
                       , plot=0 ){
  # MDPREF with Latent Class
  # Shin-ichi Mayekawa
  #
  # mdprefml modified: 20150504
  # 20150504,05,06
  # expanded data matrix version: 20150507,08
  # use of locsubj: 20150509
  # rho prior: 20150809
  # initial values: 20150510
  # npp corrected: 20150510
  # treatment of large n in lmlhmdpref: 20150515 DNC
  # reflect right and left: 20150523
  # plot: 20150423
  # error check: 20150608
  #
  #
  # Args:
  #
  #   (f, n, ij)    triplet of vectors
  #                 f[k] is the # of times that object ij[k,1]
  #                 is prefered to object ij[k,2] out of n[k] comparisons.
  #                 n can be a scalar.
  #
  #                 Elements of ij do not have to be unique.
  #                 ij can be a subset of all the possible pairs.
  #
  #   subj          vector to indicate subjects membership
  #
  #   ndim          # of dimensions
  #   nclass        # of latent classes
  #
  #   alpha         hyperparameter of Dirichlet for rho
  #
  #   init          = 0  to use random initial
  #                 = 1  to use Case V, PCA and hclust
  #                 = 2  to use MDPREF-ML, PCA and hclust
  #
  #   (X, B, rho)   initial values or NULL
  #                 If any, X and B must be given together.
  #                 (X and B) can be a result of MDPREF-ML.
  #                 This has priority over init.
  #
  #   minp          minimum value of probability
  #   lmax          max # of iterations
  #   eps           eps for relative improvement of log likelihood
  #   lmax2         max # of inner iterations
  #
  #
  #
  # Values:
  #
  #    X            nstot x ndim    matrix of centered scale values
  #    B            nsubj x nclass  matrix of subject vectors
  #    Gi           nstot x nsubj   Gi = X B'  (scale value for each subject)
  #
  #    sname_all    list of sorted unique elements of ij.
  #                 rowname(X)=sname_all
  #
  #         If Xall contains the scale values of the larger set of stimuli
  #         Xall[sname_all,]=X will put the result in it.
  #
  # Needs:
  #
  #  logistic, logit, matSwp, princ, new_name (naturalsort)
  #
  #  library(naturalsort)
  #


  lmlhmdpref <- function( f, n, ij, G, X, B, rho, alpha ){
   # evaluation of log marginal likelihood of MDPREF
   # Shin-ichi Mayekawa
   # 20150504,05,06
   # matrix version: 20150507,08
   # hyperparameter: 20150509
   # treatment of large n: 20150515 DNC
   # logistic -> logistic1: 20150629
   #
   # Args:
   #
   #       (f, n)    matrix of npairtot x nsubj
   #        ij       not used
   #       G         npairtot x nstot
   #       X         nstot x ndim
   #       B         nclass x ndim
   #       rho       nclass x 1
   #

   nsubj=ncol(f)
   nclass=nrow(B)
   ndim=ncol(X)
   nstot=nrow(X)
   npairtot=nstot*(nstot-1)/2
   P=logistic1(G%*%X%*%t(B))
   H=matrix(0,nsubj,nclass)
   sae=0
   for( s in 1:nsubj ){
    fs=f[,s]; ns=n[,s]

    temp=fs*log(P)+(ns-fs)*log(1-P)
    adjust_ex=max(colSums(temp))-700
    sae=sae+adjust_ex
    H[s,]=exp( colSums(temp) -adjust_ex )*rho
   }
   Hsum=rowSums(H)
   #  Hsum[Hsum < 1e-307]=1e-307
   H=(1/Hsum)*H

   lmlh0=sum(log(Hsum)) + sae
   lmlh=lmlh0 + sum((alpha-1)*rho)

   # calculation of nh and fh

   fh=f%*%H
   nh=n%*%H

   res=list( lmlh=lmlh, lmlh0=lmlh0, H=H, nh=nh, fh=fh )
   return( res )

  } # end of lmlhmdpref


  llhmdpref2 <- function( f, n, G, X, B ){
   # evaluation of log likelihood of cv
   # Shin-ichi Mayekawa
   # 20150423
   # matrix version: 20150508
   # logistic -> logistic1: 20150629
   #
   # Args:
   #

   P=logistic1(G%*%X%*%t(B))
   llh=sum( f*log(P)+(n-f)*log(1-P) )
   return( llh )

  } # end of llhmdpref2


  llhc5 <- function( f, n, G, x ){
   # evaluation of log likelihood of c5
   # Shin-ichi Mayekawa
   # 20150419cot
   # logistic -> logistic1: 20150629
   #
   # Args:
   #
   #  z = G x  and  p = logistic( z )
   #

   p=logistic1(G%*%x)
   llh=sum(f*log(p)+(n-f)*log(1-p))
   return( llh )

  } # end of llhc5


  getbs <- function( fs, ns, GX, bs, print=0 ){
   # estimate bs given X
   # Shin-ichi Mayelawa
   # 20150426
   # estbs renamed as getbs: 20150506
   # logistic -> logistic1: 20150629
   #

   llhbsp=llhc5( fs, ns, GX, bs )
   llhbs=llhbsp

   # first derivative of llh
   zs=GX%*%bs
   ps=logistic1(zs)
   dllhdb=1.7*t(GX)%*%(fs-ns*ps)
   maxagbs=max(abs(dllhdb))

   # Fisher info mat
   H=1.7^2*t(GX)%*%diag(c(ns*ps*(1-ps)))%*%GX
   # direction
   d=matSwp(H)%*%dllhdb

   # halving
   step=1
   for( lll in 1:20 ){
    bs1=bs+step*d
    llhbs1=llhc5( fs, ns, GX, bs1 )
    if( llhbs1 > llhbs ) break
    step=step/2
   }

   # update bs assuming llhbs1 > llhbs
   bs=bs1
   llhbs=llhbs1

   llhbsimp=llhbs-llhbsp
   llhbsimpr=llhbsimp/abs(llhbsp)

   if( print >= 3 ){
    Print( ll, lll, llhbs, llhbsimp, maxagbs )
   }

   return( list(bs=bs, maxagbs=maxagbs, llhbs=llhbs) )

  } # end of getbs


  getX <- function( f, n, ij, G, X, B, print=0 ){
   # estimate X given B
   # Shin-ichi Mayelawa
   # 20150426,0506
   # estX renamed as getbs: 20150506
   # matrix version: 20150508
   #

   notpt=nrow(X)
   nclass=nrow(B)
   llhx=llhmdpref2( f, n, G, X, B )
   llhxp=llhx
   x=matrix(X,,1)

   # first derivative and Fisher info mat
   dllhdx=0
   H=0
   for( s in 1:nclass ){
    ns=n[,s]; fs=f[,s]
    bs=t(B[s,,drop=0]) # this is nr x 1
    ps=logistic1(G%*%X%*%bs)
    ttbsG=t(t(bs)%x%G)
    dllhdx=dllhdx+ttbsG%*%(fs-ns*ps)
    H=H+ttbsG%*%(diag(c(ns*ps*(1-ps)))%*%t(ttbsG))
   } # end of subj
   dllhdx=1.7*dllhdx
   H=(1.7^2)*H
   maxagX=max(abs(dllhdx))

   # direction
   d=matSwp(H)%*%dllhdx

   # halving
   step=1
   for( lll in 1:10 ){
    x1=x+step*d
    X1=matrix(x1,nstot)
    # center X1 before evaluating llh
    X1=scale(X1, center=TRUE, scale=FALSE)
    llhx1=llhmdpref2( f, n, G, X1, B )
    if( llhx1 > llhx ) break
    step=step/2
   }

   # update X assuming llhx1 > llhx
   X=X1
   llhx=llhx1

   llhximp=llhx-llhxp
   llhximpr=llhximp/abs(llhxp)

   if( print >= 3 ){
    Print( ll, lll, llhx, llhximp, maxagX )
   }

   return( list(X=X, maxagX=maxagX, llhx=llhx) )

  } # end of getX


  if( length(n) == 1 ) n=rep(n,length(f))

  # error check
  if( !is.vector(f) | !is.vector(n) | !is.matrix(ij) ){
   cat("\n\nerror1:(mdprefml) f must be a vector, ij must be a matrix. \n\n")
   return()
  }


  # rename ij
  #sname_all=sort(unique(c(ij)))
  #o2n=compress_id( ij )$o2n
  #ij=matrix( o2n[ij],,2 )
  nnn=new_name( ij, sort=2, print=0 )
  ij=nnn$new

  # reflect right and left if left < right.
  loc=which(ij[,1] < ij[,2])
  ij[loc,]=cbind(ij[loc,2],ij[loc,1])
  f[loc]=n[loc]-f[loc]

  n2o=nnn$n2o
  sname_all=n2o$old
  nstot=length(unique(c(ij)))
  rm(nnn)

  # From here, ij contains numeric ids of nstot stimuli.

  # const
  subjname=unique(subj)      # not sorted
  nsubj=length(subjname)
  npair=numeric(nsubj)
  npairfull=numeric(nsubj)
  nstim=numeric(nsubj)
  npairu=numeric(nsubj)
  ntot=numeric(nsubj)
  pmiss=numeric(nsubj)
  npairtot=nstot*(nstot-1)/2
  sname=NULL
  locsubj=NULL
  locmiss=vector("list",nsubj)
  if( length(n) == 1 ) n=rep(n,length(f))
  for( s in 1:nsubj ){
   loc=which(subj == subjname[s])
   locsubj[[s]]=loc
   fs=f[loc]; ns=n[loc]; ijs=ij[loc,]
   npair[s]=length(fs)
   sname[[s]]=sort(unique(c(ijs)))
   if( length(setdiff(1:nstot,sname[[s]])) > 0 )
    locmiss[[s]]=setdiff(1:nstot,sname[[s]])
   nstim[s]=length(sname[[s]])
   npairu[s]=nrow(unique(ijs))
   npairfull[s]=nstim[s]*(nstim[s]-1)/2
   ntot[s]=sum(ns)
   pmiss[s]=1-npairu[s]/npairfull[s]
  }

  dimname=paste("d",1:ndim,sep="")
  classname=paste("class",1:nclass,sep="")

  nparam=(nstot-1)*ndim + nclass*ndim + nclass-1  -
   ndim*(ndim+1)/2 - ndim*(ndim-1)/2
  if( estX == 0 & estB == 1 ) nparam=nclass*ndim + nclass-1
  if( estX == 1 & estB == 0 ) nparam=(nstot-1)*ndim + nclass-1
  if( estrho == 0 )nparam=nparam-(nclass-1)

  # prior constant
  if( length(alpha) == 1 ) alpha=rep(alpha,nclass)

  # initial
  if( !is.null(B) ){
   initin=1
   if( nrow(B) == nclass ) initin=2
   if( is.null(rho) ) rho=rep(1/nclass,nclass)
  }
  else{
   initin=0
  }



  # big G matrix
  GG=matrix(0,npairtot,nstot)
  ijij=matrix(0,npairtot,2)
  k=0
  for( i in 2:nstot )
   for( j in 1:(i-1) ){
    k=k+1
    GG[k,i]=1; GG[k,j]=-1
    ijij[k,1]=i; ijij[k,2]=j
   }
  ijtot=matrix(rep(t(ijij),nclass),,2,byrow=1)
  subjtot=c( sapply(1:nclass,rep,npairtot) )

  # big data vector
  ff=matrix(0,npairtot,nsubj)
  nn=matrix(0,npairtot,nsubj)
  colnames(ff)=subjname; colnames(nn)=subjname
  for( s in 1:nsubj ){
   locs=locsubj[[s]]
   fs=f[locs]; ns=n[locs]; ijs=ij[locs,,drop=0]
   # Print(s,sort(unique(c(ijs))),sname_all[sort(unique(c(ijs)))])
   for( k in 1:length(fs) ){
    is=ijs[k,1]; js=ijs[k,2]
    kk=(is-1)*(is-2)/2+js
    nn[kk,s]=nn[kk,s]+ns[k]; ff[kk,s]=ff[kk,s]+fs[k]
    # Print(s,k,is,js,kk)
   }
  }

  npp=numeric(nsubj)
  for( s in 1:nsubj ){
   ns=nn[,s]; ns=ns[ns>0]
   npp[s]=mean(ns)
  }


  if( print > 0 ){
   cat("\n\nMDPREF with Latent Class\n\n")
   # Print(nsubj, npair, npairu, ns, npairfull, pmiss, ntot, npp, locsubj)
   # Print(sname)
   cat(" total # of stimuli used =",nstot,"\n")
   cat(" # of subjects =", nsubj,"\n")
   cat(" # of dimensions =", ndim,"\n")
   cat(" # of latent classes =", nclass,"\n\n")
   cat(" # of pairs observed =", npair, "\n")
   cat(" # of unique pairs observed =", npairu, "\n")
   cat(" # of stimuli =", nstim, "\n")
   cat(" # of full pairs =", npairfull,"\n")
   cat(" proportion missing =", pmiss, "\n")
   cat(" total # of comparisons =", ntot, "\n")
   cat(" # of trials per pair =", npp, "\n")
   cat("\n # of parameters to be estimated =", nparam,"\n")
   cat(" initin =", initin, ",  init =", init, "\n")

   cat("\n estimation of X =", estX, ",  estimation of B =", estB)
   cat(",  estimation of rho =", estrho,"\n")
   if( estrho == 1 )
    cat("  hyperparameter = ", alpha, "\n")
   cat("\n max # of iterations =", lmax, ",  eps =", eps, "\n")
   cat(" max # of inner iterations =", lmax2, ",  eps =", eps, "\n")
   cat(" Minimum value of prob is set equal to", minp, " .\n")
  }


  if( print >= 4 ){
   cat("\n Expanded Data Matrix\n")
   Print( ijij, nn, ff )
  }


  # initial values
  if( !initin  &  init == 1 ){

   # initial by case five solution
   Gi=matrix(0,nstot,nsubj)
   for( s in 1:nsubj ){
    G=matrix(0,npair[s],nstot)
    ijs=ij[locsubj[[s]],]
    for( k in 1:npair[s] ){
     G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1
    }
    ps=f[locsubj[[s]]]/n[locsubj[[s]]]
    zs=logit1(ps, 0.1)
    x=matSwp(t(G)%*%G)%*%t(G)%*%zs
    x[locmiss[[s]]]=NA
    Gi[,s]=x
   }
   Gi=scale(Gi, center=TRUE, scale=FALSE)

   # expand or shrink x0 to find a better initial
   for( s in 1:nsubj ){
    # subsetting
    locs=locsubj[[s]]
    ijs=ij[locs,]; fs=f[locs]; ns=n[locs]; npairs=nrow(ijs)
    G=matrix(0,npairs,nstot);
    for( k in 1:npairs ){ G[k,ijs[k,1]]=1; G[k,ijs[k,2]]=-1 }

    x0=Gi[,s]
    x0[locmiss[[s]]]=0
    llh00=llhc5( fs, ns, G, x0 )
    llh0=llh00
    ddd=1
    for( dd in c(3, 2, 1.5, 0.8, 0.6, 0.4, 0.2, 0.1, 0.05) ){
     llh=llhc5( fs, ns, G, dd*x0 )
     if( llh > llh0 ){
      ddd=dd
      llh0=llh
     }
    }
    Gi[,s]=ddd*x0
    Gi[locmiss[[s]],s]=NA
    if( print >= 5 ) Print( s, ddd, llh0, llh00 )
   }

   # initial X and B by pca of Gi=X B'
   res=princ( Gi, ndim, estmu=1, maxiter=1000, eps=1e-4, init=2, print=0 )
   X=res$F; B=res$A
   rm(res,G,Gi)
   rownames(X)=sname_all
   colnames(X)=dimname
   colnames(B)=dimname

  } # end of !initin and init == 2
  else if( !initin  &  init == 2 ){

   temp=mdprefml( f, n, ij, subj, print=0, eps=1e-5, lmax2=1 )
   X=temp$X; B=temp$B
   rm(temp)

  } # end of !initin and init == 2

  rho=rep(1/nclass,nclass)

  if( initin <= 1 ){
   # initial class means
   cluster=cutree( hclust( dist(B), method="ward.D2" ), nclass )
   B=t( sapply(
    tapply( 1:nsubj, cluster, c ), function(s) colSums(B[s,,drop=0]) ) )
   rownames(B)=classname
   colnames(B)=dimname
   rm(cluster)
   # rho=as.vector( table(cluster) )/nsubj
  }
  Gi=GG%*%X%*%t(B)

  # initial ff and nn
  temp=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )
  lmlh0=temp$lmlh
  fh=temp$fh
  nh=temp$nh
  H=temp$H

  if( print >= 2 ){
   cat("\n Initial Parameter Values\n")
   Print(X,B,rho,lmlh0)
  }



  # main iteration starts here

  lmlh=lmlh0
  lmlhp=lmlh0
  maxagX=0
  convXB=0

  for( llll in 1:lmax ){

   # update B given X

   maxagB=0
   for( q in 1:nclass ){
    bq=B[q,]
    for( ll in 1:lmax2 ){
     res=getbs( fh[,q], nh[,q], GG%*%X, bq )
     bq=res$bs
     maxagB=max(maxagB,res$maxagbs)
    }
    B[q,]=bq
    if( print >= 5 ){
     lmlh=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )$lmlh
     Print("after B",q,lmlh)
    }
   } # end of class

   temp=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )
   lmlh=temp$lmlh
   # Print("after B  ", lmlh)


   if( estX ){
    # update X given B
    for( ll in 1:lmax2 ){
     res=getX( fh, nh, ijij, GG, X, B )
     X=res$X
     rownames(X)=sname_all
     # Print("X ", ll, X)
    }
    maxagX=res$maxagX

    if( print >= 5 ){
     temp=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )
     lmlh=temp$lmlh
     Print("after X  ", lmlh)
    }
   }


   if( llll%%10 == 0 ){
    # normalize X and B
    # make X orthonormal
    ev=eigen( t(X)%*%X/nstot )
    evamhalf=diag(sqrt(1/ev$values)); eve=ev$vectors
    X=X%*%eve%*%evamhalf
    B=B%*%eve%*%diag(sqrt(ev$values))
    # make B orthogonal
    eve=eigen( t(B)%*%B )$vector
    B=B%*%eve
    X=X%*%eve
    #Print("normalized")
   }


   if( estrho & convXB ){
    # update rho
    rho=colSums(H)+alpha-1
    rho=rho/sum(rho)
    if( print >= 5 ){
     temp=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )
     lmlh=temp$lmlh
     Print("after rho  ", lmlh)
    }
   }


   # check convergence and update fh and nh
   temp=lmlhmdpref( ff, nn, ijij, GG, X, B, rho, alpha )
   lmlh=temp$lmlh
   lmlh0=temp$lmlh0
   fh=temp$fh
   nh=temp$nh
   H=temp$H
   rownames(H)=subjname; colnames(H)=classname

   lmlhimp=lmlh-lmlhp
   lmlhimpr=lmlhimp/abs(lmlhp)

   if( print >= 2 ){
    Print( llll, lmlh, lmlhimpr, lmlh0 )
   }

   if( lmlhimpr <= eps ){
    if( convXB == 1 ){
     if( print > 0 ) cat("\n mdprefml iteration converged.\n" )
     break
    }
    else{
     convXB=1
     cat("\n convXB is set equal to 1.\n\n")
    }
   }

   lmlhp=lmlh

  } # end of llll loop


  # normalize X and B so that X'X = nstot I  and B'B = diag.
  # make X orthonormal
  X=scale(X,center=TRUE,scale=FALSE)
  ev=eigen( t(X)%*%X/nstot )
  evamhalf=diag(sqrt(1/ev$values)); eve=ev$vectors
  X=X%*%eve%*%evamhalf
  B=B%*%eve%*%diag(sqrt(ev$values))
  # make B orthogonal
  eve=eigen( t(B)%*%B )$vector
  B=B%*%eve
  X=X%*%eve
  # make B positive
  bs=colSums(B>0) < nclass/2
  B[,bs]=-B[,bs]
  X[,bs]=-X[,bs]

  rownames(X)=sname_all
  colnames(X)=dimname
  rownames(B)=classname
  colnames(B)=dimname

  # error variance etc
  Gi=X%*%t(B)
  md=mandd(Gi)
  DispGi=md$cov
  stdGi=md$std
  CorGi=diag(1/stdGi)%*%DispGi%*%diag(1/stdGi)
  dimnames(CorGi)=dimnames(DispGi)
  BB=t(B)%*%B
  rcont=diag(BB)/sum(diag(BB))
  names(rcont)=dimname

  aic=-2*(lmlh-nparam)
  aic0=-2*(lmlh0-nparam)
  bic=-2*lmlh+nparam*log(sum(ntot))
  bic0=-2*lmlh0+nparam*log(sum(ntot))

  class=max.col(H)

  if( print > 0 ){
   cat("\nMDPREF with Latent Class\n")
   Print( lmlh, lmlh0, maxagX, maxagB )
   Print( nparam, aic, aic0, bic, bic0 )
   cat("\n")
   Print( X, B, rho, rcont, fmt="6.3" )
   cat("\n Scale Values for each class\n")
   Print(Gi, fmt="6.3")
   Print(DispGi, BB, fmt="6.3")
   Print(CorGi, fmt="6.3")
   if( print >= 3 ){
    cat("\n Posterior Distribution of Class Indicator for each subject\n")
    Print(H, class)
   }
  }

  if( plot >=  1 ){
   title=paste("Biplot:  # of params = ", nparam
               ,",   lmlh = " ,formatC(lmlh,8,format="fg")
               ,",   aic = ",formatC(aic,8,format="fg")
               ,",   bic = ", formatC(bic,8,format="fg"), sep="")
   biplot( X,B, main=title)
  }

  res=list( X=X, B=B, rho=rho, sname_all=sname_all, sname=sname, n2o=n2o
            , npair=npair, npairu=npairu, ntot=ntot
            , lmlh=lmlh, lmlh0=lmlh0, maxagX=maxagX, maxagB=maxagB
            , aic=aic, aic0=aic0, bic=bic, bic0=bic0, nparam=nparam
            , Gi=Gi, DispGi=DispGi, iter=llll, eps=eps, H=H, class=class  )
  return( res )

 } # end of mdprefmx





#' Genation of Test Data Frame
#'
#' Pleae Do NOT Run This.
#' Just type gendatamdpref to see how I generated test data frame.
#'
#' @export

gendatamdpref <- function(){
 # Data Generation for mdprefml and mdprefmx
 # Shin-ichi Mayekawa
 # 20150418,0606
 #


 # initialization of random number seed
 cat("# initialization of random number seed.\n")
 seed=093245+5
 set.seed(seed)

 # common # of trials per pair to be used in rbin
 cat("# common # of trials per pair to be used in rbin.\n")
 ncommon=1

 # percentage of missing data
 cat("# percentage of missing data .\n")
 pmiss=0.2



 # Stimulus configuration matrix X: np x nr
 cat("# Stimulus configuration matrix X: np x nr .\n")
 X=matrix(c(
  -1,  1,  0,  1,  1,    1, -1,   0, 0,  0, 1, 0, -1, -1, 0, -1, 1, -1
  , -.5,  .5, .5,  .5, -.5,  -.5, .5,  -.5, -2,  0, 2,  0
 ),,2,byrow=1)


 # Subject vector matrix Beta: nK x nr
 # Generated via angles from x-axis as
 # Beta=cbind(cos(theta),sin(theta))
 cat("# Subject vector matrix Beta: nK x nr .\n")
 cat("# Generated via angles from x-axis as. .\n")
 cat("# Beta=cbind(cos(theta),sin(theta)) .\n")
 theta=c(45,30,0,-30,-45)/360*2*pi
 # theta=c(theta,theta-5/360*2*pi)
 cat("# theta=c(theta,theta-5/360*2*pi) .\n")
 Beta=cbind(cos(theta),sin(theta))

 ###########################################################################
 # From here, it's automatic.
 ###########################################################################
 cat("# From here, it's automatic. .\n")

 # Size of matrices
 cat("# Size of matrices .\n")
 np=nrow(X)      # of stimuki
 nr=ncol(X)      # of dimensions
 nK=nrow(Beta)   # of subjects

 # total # of pairs
 cat("# total # of pairs .\n")
 npair=np*(np-1)/2


 # N matrix for each pair and subject: npair x nK
 cat("# N matrix for each pair and subject: npair x nK .\n")
 N=matrix(ncommon,npair,nK)


 # Scale score matrix   np x nK
 cat("# Scale score matrix   np x nK .\n")
 Gi=X%*%t(Beta)

 # Standardize it
 cat("# Standardize it .\n")
 std=sqrt(apply(Gi,2,var)*(np-1)/np)
 Gi=Gi%*%diag(1/std)

 # The design matrix G:  npair x np
 # The index matrix IJ: npair x 2
 cat("# The design matrix G:  npair x np .\n")
 cat("# The index matrix IJ: npair x 2 .\n")
 G=matrix(0,npair,np)
 IJ=matrix(c(0,0),npair,2)
 k=0
 for( i in 2:np )
  for( j in 1:(i-1) ){
   k=k+1
   G[k,i]=1; G[k,j]=-1
   IJ[k,1]=i; IJ[k,2]=j
  }


 # Matrix of paird difference for each subject Y: npair x nK
 cat("# Matrix of paird difference for each subject Y: npair x nK .\n")
 Y=G%*%Gi

 # Probability matrix: same size as Y
 cat("# Probability matrix: same size as Y .\n")
 P=logistic(Y)

 # Frequency matrix from rbin
 cat("# Frequency matrix from rbin .\n")
 F=matrix(0,npair,nK)
 for( k in 1:nK ){
  # F[,k]=rbin(N[,k],P[,k])
  F[,k]==mapply( function(size,prob){ rbinom(1,size,prob) }, N[,k], P[,k] )
 }

 # F and N are npair x nK matrices.
 # From here, we roll-out those two matrices to form vectors
 # create a vector subj consisting of subject name.
 # Also, we create ij index matrix ( npair x 2 ) for each subject
 # in order to indicate the observed pairs.
 cat("# F and N are npair x nK matrices.
     # From here, we roll-out those two matrices to form vectors
     # create a vector subj consisting of subject name.
     # Also, we create ij index matrix ( npair x 2 ) for each subject
     # in order to indicate the observed pairs. .\n")

 # roll-out
 cat("# Roll-out F and N to create f and n. And expand IJ to form ij.\n")
 f=matrix(F,,1)
 ij=cbind(rep(IJ[,1],nK),rep(IJ[,2],nK))
 n=matrix(N,,1)
 subj=unlist(sapply(1:nK,rep,npair,simplify=FALSE))

 # Change subject #1 to #91 for debuging purpose
 cat("# Change subject #1 to #91 for debuging purpose .\n")
 subj[subj==1]=91

 # Make subj character for debuging purpose
 cat("# Make subj character for debuging purpose .\n")
 subj=paste("Sbj",subj,sep="")


 # Create complete data (no missing pairs)
 cat("# Create complete data (no missing pairs) .\n")
 # data_comp=data.frame( subj=subj, n=n, f=f, ij=ij, stringsAsFactors=0  )
 # use_data(data_comp, overwrite=1)



 # Missing data
 cat("# Missing data .\n")
 locobs=sort(sample(1:length(f), length(f)*(1-pmiss)))

 # Pick up observed pair only
 cat("# Pick up observed pair only .\n")
 f=f[locobs]
 ij=ij[locobs,]
 n=n[locobs]
 subj=subj[locobs]


 # Systematic missing
 cat("# Systematic missing .\n")
 loc1=NULL; loc2=NULL; loc3=NULL
 loc1=intersect( which(subj==1), union(which(ij[,1]==1), which(ij[,2]==1)) )
 loc2=intersect( 1:length(ij), union(which(ij[,1]==2), which(ij[,2]==2)) )
 loc3=intersect( which(subj==3)
                 , union(which(ij[,1]%in%c(6,8)), which(ij[,2]%in%c(6,8))) )
 loc12=c(loc1,loc2,loc3)
 ij=ij[-loc12,]
 f=f[-loc12]
 n=n[-loc12]
 subj=subj[-loc12]


 # Switching i and j for about half of pairs
 cat("# Switching i and j for about half of pairs .\n")
 loc=sort(sample(1:length(f), 0.5*length(f)))
 ij[loc,]=ij[loc,2:1]
 f[loc]=n[loc]-f[loc]

 # Save numeric ij
 cat("# Save numeric ij .\n")
 ij_num=ij
 ij=matrix(paste("S",ij,sep="",colapse=""),nrow(ij))

 # Create complete data (no missing pairs)
 cat("# Create incomplete data (with missing pairs) .\n")
 # data_miss=data.frame( subj=subj, n=n, f=f, ij=ij, stringsAsFactors=0  )
 # use_data(data_miss, overwrite=1)

} # end of gendatamdpref


